/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Content Preferences (cpref).
 *
 * The Initial Developer of the Original Code is Mozilla.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Myk Melez <myk@mozilla.org> (original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const Ci = Components.interfaces;
const Cc = Components.classes;
const Cr = Components.results;

function ContentPreferencesService() {}

ContentPreferencesService.prototype = {
  interfaces: [Ci.mozIContentPreferencesService, Ci.nsISupports],

  QueryInterface: function (iid) {
    if (!this.interfaces.some( function(v) { return iid.equals(v) } ))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  },

  database: null,

  _init: function cps_init() {
    // Set up the database.
    this.initDatabase();
  },

  destroy: function cps_destroy() {
    // XXX Is this sufficient to shut down the database connection?
    this.database = null;
  },

  __getSitePrefsStatement: null,
  get _getSitePrefsStatement() {
    if (!this.__getSitePrefsStatement) this.__getSitePrefsStatement =
      this.createStatement(
        "SELECT prefs.id AS id, keys.name AS key, prefs.value AS value \
         FROM sites LEFT JOIN prefs ON sites.id = prefs.site_id \
         LEFT JOIN keys ON prefs.key_id = keys.id \
         WHERE sites.name = :site"
      );
    return this.__getSitePrefsStatement;
  },

  __getGlobalPrefsStatement: null,
  get _getGlobalPrefsStatement() {
    if (!this.__getGlobalPrefsStatement) this.__getGlobalPrefsStatement =
      this.createStatement(
        "SELECT prefs.id AS id, keys.name AS key, prefs.value AS value \
         FROM prefs LEFT JOIN keys ON prefs.key_id = keys.id \
         WHERE site_id IS NULL"
      );
    return this.__getGlobalPrefsStatement;
  },

  /*
   * Get the set of preferences for a given site.
   * 
   */
  getPrefs: function cps_getPrefs(site, prefCount) {
    var prefs = [];

    var stmt = site ? this._getSitePrefsStatement : this._getGlobalPrefsStatement;

    stmt.reset();
    if (site)
      stmt.params.site = site;
    while(stmt.step())
      prefs.push(new ContentPref(stmt.row["id"], site, stmt.row["key"], stmt.row["value"]));
    stmt.reset();

    prefCount.value = prefs.length;
    return prefs;
  },

  _getPrefStatement: null,

  /* Get the value of a preference from the database.
   *
   * @param   site   the name of the site to which the pref applies
   * @param   key    the name of the key identifying the pref
   * @returns the value of the pref, or null if it doesn't exist
   */
  getPref: function cps_getPref(site, key) {
    if (!site)
      return this.getGlobalPref(key);

    var pref = null;

    if (!this._getPrefStatement) this._getPrefStatement =
      this.createStatement(
        "SELECT prefs.id AS id, prefs.value AS value \
         FROM prefs JOIN sites ON prefs.site_id = sites.id \
         JOIN keys ON prefs.key_id = keys.id \
         WHERE sites.name = :site AND keys.name = :key"
      );

    this._getPrefStatement.reset();
    this._getPrefStatement.params.site = site;
    this._getPrefStatement.params.key = key;
    if (this._getPrefStatement.step())
      pref = new ContentPref(this._getPrefStatement.row["id"],
                             site,
                             key,
                             this._getPrefStatement.row["value"]);
    this._getPrefStatement.reset();

    return pref;
  },

  _getGlobalPrefStatement: null,

  /* Get the value of a global preference from the database.  Global prefs
   * Are stored in the same table as site-specific prefs, but the value
   * of their "site_id" field is null.
   *
   * @param    key    the name of the key identifying the pref
   * 
   * @returns  the value of the pref, or null if it doesn't exist
   */
  getGlobalPref: function cps_getGlobalPref(key) {
    var pref = null;

    if (!this._getGlobalPrefStatement) this._getGlobalPrefStatement =
      this.createStatement(
        "SELECT prefs.id AS id, prefs.value AS value \
         FROM prefs JOIN keys ON prefs.key_id = keys.id \
         WHERE site_id IS NULL AND keys.name = :key"
      );

    this._getGlobalPrefStatement.reset();
    this._getGlobalPrefStatement.params.key = key;
    if (this._getGlobalPrefStatement.step())
      pref = new ContentPref(this._getGlobalPrefStatement.row["id"],
                             null,
                             key,
                             this._getGlobalPrefStatement.row["value"]);
    this._getGlobalPrefStatement.reset();

    return pref;
  },

  /* Set the pref to the given value.
   */
  setPref: function cps_setPref(site, key, value) {
    var pref = this.getPref(site, key);
    if (pref && pref.value == value)
      return;
    else if (pref)
      this.updatePref(pref, value);
    else
      this.insertPref(site, key, value);
  },

  _updatePrefStatement: null,

  updatePref: function cps_updatePref(pref, newValue) {
    if (!this._updatePrefStatement) this._updatePrefStatement =
      this.createStatement("UPDATE prefs SET value = :value WHERE id = :id");

    this._updatePrefStatement.reset();
    this._updatePrefStatement.params.id = pref.id;
    this._updatePrefStatement.params.value = newValue;
    this._updatePrefStatement.step();
    this._updatePrefStatement.reset();
  },

  _insertPrefStatement: null,

  insertPref: function cps_insertPref(siteName, keyName, value) {
    if (!this._insertPrefStatement) this._insertPrefStatement =
      this.createStatement("INSERT INTO prefs (site_id, key_id, value) \
                  VALUES (:site_id, :key_id, :value)");

    var site = siteName ? this.getSite(siteName) || this.insertSite(siteName) : null;
    var key = this.getKey(keyName) || this.insertKey(keyName);

    this._insertPrefStatement.reset();
    this._insertPrefStatement.params.site_id = site.id;
    this._insertPrefStatement.params.key_id = key.id;
    this._insertPrefStatement.params.value = value;
    this._insertPrefStatement.step();
    this._insertPrefStatement.reset();
  },

  _getSiteStatement: null,
  
  getSite: function cps_getSite(name) {
    var site = null;

    if (!this._getSiteStatement) this._getSiteStatement =
      this.createStatement("SELECT id from sites WHERE name = :name");

    this._getSiteStatement.reset();
    this._getSiteStatement.params.name = name;
    if (this._getSiteStatement.step())
      site = { id: this._getSiteStatement.row["id"] };
    this._getSiteStatement.reset();
    
    return site;
  },

  _insertSiteStatement: null,

  insertSite: function cps_insertSite(name) {
    var site = { name: name };

    if (!this._insertSiteStatement) this._insertSiteStatement =
      this.createStatement("INSERT INTO sites (name) VALUES (:name)");

    this._insertSiteStatement.reset();
    this._insertSiteStatement.params.name = name;
    this._insertSiteStatement.step();
    site.id = this.database.lastInsertRowID;
    this._insertSiteStatement.reset();

    return site;
  },

  _getKeyStatement: null,
  
  getKey: function cps_getKey(name) {
    var key = null;

    if (!this._getKeyStatement) this._getKeyStatement =
      this.createStatement("SELECT id from keys WHERE name = :name");

    this._getKeyStatement.reset();
    this._getKeyStatement.params.name = name;
    if (this._getKeyStatement.step())
      key = { id: this._getKeyStatement.row["id"] };
    this._getKeyStatement.reset();
    
    return key;
  },

  _insertKeyStatement: null,

  insertKey: function cps_insertKey(name) {
    var key = { name: name };

    if (!this._insertKeyStatement) this._insertKeyStatement =
      this.createStatement("INSERT INTO keys (name, type) VALUES (:name, 1)");

    this._insertKeyStatement.reset();
    this._insertKeyStatement.params.name = name;
    this._insertKeyStatement.step();
    key.id = this.database.lastInsertRowID;
    this._insertKeyStatement.reset();

    return key;
  },

  // storage stuff

  _dbService: null,
  get dbService() {
    if (!this._dbService)
    this._dbService = Cc["@mozilla.org/storage/service;1"].
           getService(Ci.mozIStorageService);
    return this._dbService;
  },

  schema: {
    sites:  "id       INTEGER PRIMARY KEY, \
             name     STRING NOT NULL",

    keys:   "id       INTEGER PRIMARY KEY, \
             name     STRING NOT NULL, \
             type     INTEGER NOT NULL",

    prefs:  "id       INTEGER PRIMARY KEY, \
             site_id  INTEGER REFERENCES sites(id), \
             key_id   INTEGER NOT NULL REFERENCES keys(id), \
             value    STRING NOT NULL"
  },

  createStatement: function cps_createStatement(sqlString) {
    var statement = this.database.createStatement(sqlString);
    var wrappedStatement = Cc["@mozilla.org/storage/statement-wrapper;1"].
                 createInstance(Ci.mozIStorageStatementWrapper);
    wrappedStatement.initialize(statement);
    return wrappedStatement;
  },

  initDatabase: function cps_initDatabase() {
    var dirService = Cc["@mozilla.org/file/directory_service;1"].
             getService(Ci.nsIProperties);
    var file = dirService.get("ProfD", Ci.nsIFile);
    file.append("content-prefs.sqlite");

    try {
      this.database = this.dbService.openDatabase(file);
    } catch (ex) {
      if (ex.result == Cr.NS_ERROR_FILE_CORRUPTED) {
        // Delete and recreate the file.
        // XXX Shouldn't we make a copy in case there's some way
        // to recover the data?
        file.remove(false);
        this.database = this.dbService.openDatabase(file);
      } else {
        throw ex;
      }
    }
  
    for (var table in this.schema) {
      try {
        this.database.createTable(table, this.schema[table]);
      } catch (ex) {
        // createTable will fail if the table already exists, which is fine.
        // XXX Make sure the failure is really because the table already exsits.
      }
    }
  }
};

function ContentPref(id, site, key, value) {
  if (typeof id != "undefined" && id != null)
    this.id = id;
  if (typeof site != "undefined" && site != null)
    this.site = site;
  if (typeof key != "undefined" && key != null)
    this.key = key;
  if (typeof value != "undefined" && value != null)
    this.value = value;
}

ContentPref.prototype = {
  interfaces: [Ci.mozIContentPref, Ci.nsISupports],

  QueryInterface: function (iid) {
    if (!this.interfaces.some( function(v) { return iid.equals(v) } ))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  },

  id: null,
  site: null,
  key: null,
  value: null
};

var gModule = {
  registerSelf: function(componentManager, fileSpec, location, type) {
    componentManager = componentManager.QueryInterface(Ci.nsIComponentRegistrar);
    
    for (var key in this._objects) {
      var obj = this._objects[key];
      componentManager.registerFactoryLocation(obj.CID,
                                               obj.className,
                                               obj.contractID,
                                               fileSpec,
                                               location,
                                               type);
    }
  },
  
  unregisterSelf: function(componentManager, fileSpec, location) {},

  getClassObject: function(componentManager, cid, iid) {
    if (!iid.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
  
    for (var key in this._objects) {
      if (cid.equals(this._objects[key].CID))
      return this._objects[key].factory;
    }
    
    throw Components.results.NS_ERROR_NO_INTERFACE;
  },
  
  _objects: {
    service: {
      CID        : Components.ID("{caecdbe8-826d-4920-9c7c-c0b99bd282ee}"),
      contractID : "@mozilla.org/content-prefs/service;1",
      className  : "Content Preferences Service",
      factory    : ContentPreferencesServiceFactory = {
                     createInstance: function(aOuter, aIID) {
                       if (aOuter != null)
                         throw Components.results.NS_ERROR_NO_AGGREGATION;
                       var svc = new ContentPreferencesService();
                       svc._init();
                      return svc.QueryInterface(aIID);
                     }
                   }
    }
  },
  
  canUnload: function(componentManager) {
    return true;
  }
};

function NSGetModule(compMgr, fileSpec) {
  return gModule;
}
