/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Content Preferences (cpref).
 *
 * The Initial Developer of the Original Code is Mozilla.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Myk Melez <myk@mozilla.org> (original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

if (typeof Ci == "undefined") {
  var Ci = Components.interfaces;
  var Cc = Components.classes;
  var Cr = Components.results;
}

var gTextZoomPrefHandler = {
  key: "text.zoom",

  // content preferences service
  __cps: null,
  get _cps() {
    if (!this.__cps)
      this.__cps = Cc["@mozilla.org/content-prefs/service;1"].
                   getService(Ci.mozIContentPreferencesService);
    return this.__cps;
  },

  // Interfaces this component implements.
  interfaces: [Ci.nsIDOMEventListener, Ci.nsISupports],
  
  // nsISupports
  
  QueryInterface: function tzpf_QueryInterface(iid) {
    if (!this.interfaces.some( function(v) { return iid.equals(v) } ))
      throw Cr.NS_ERROR_NO_INTERFACE;
    return this;
  },

  init: function tzpf_init() {
    // Watch for View > Text Size > Increase/Decrease/Normal commands
    // so we can update the database and the sidebar UI accordingly.
    document.getElementById("cmd_textZoomEnlarge")
            .addEventListener("command", this, false);
    document.getElementById("cmd_textZoomReduce")
            .addEventListener("command", this, false);
    document.getElementById("cmd_textZoomReset")
            .addEventListener("command", this, false);
  },

  // nsIDOMEventListener

  handleEvent: function tzpf_handleEvent(event) {
    switch(event.target) {
    case document.getElementById("cmd_textZoomEnlarge"):
    case document.getElementById("cmd_textZoomReduce"):
    case document.getElementById("cmd_textZoomReset"):
      this.handleMenuChange(this.zoomManager.textZoom);
      break;
    default:
      cpref_log("unexpected event: " + event.type);
      break;
    }
  },

  _zoomManager: null,
  get zoomManager() {
    if (!this._zoomManager)
      this._zoomManager = ZoomManager.prototype.getInstance();
    return this._zoomManager;
  },

  get sidebar() {
    var sidebar = document.getElementById("sidebar");
    if (sidebar && sidebar.getAttribute("src") == "chrome://cpref/content/sidebar.xul")
      return sidebar.contentDocument;
    return null;
  },

  /*
   * Per DocumentViewerImpl::SetTextZoom in nsDocumentViewer.cpp, it looks
   * like we can set the zoom to its current value without significant impact
   * on performance, as the setting is only applied if it differs from the
   * current setting.
   *
   * And perhaps we should always set the zoom even if it were to incur
   * a performance penalty, since SetTextZoom claims that child documents
   * may have a different zoom under unusual circumstances, and it implies
   * that those child zooms should get updated when the parent zoom gets set.
   *
   * So in the methods below, when we apply new zoom values to the browser,
   * we simply set the zoom.  We don't check first to see if the new value
   * is the same as the current one.
   * 
   */

  /**
   * Apply the preference for the new location.  For performance, the content
   * prefs controller queries the database once for all available preferences
   * and passes each to its handler, so we don't need to query the database
   * ourselves, we just need to apply the value that the controller gives us
   * to the browser and then update the sidebar to reflect the change.
   *
   * @param value the pref for the new location, or null if there isn't one
   * 
   */
  handleLocationChange: function tzpf_handleLocationChange(value) {
    cpref_log("handle location change; new zoom: " + value);

    // Apply the pref to the browser.
    if (value)
      this.zoomManager.textZoom = value;
    else {
      this.zoomManager.reset();
      value = this.zoomManager.textZoom;
    }

    // Update the control in the sidebar to reflect the change.
    this.updateControl(value);
  },

  handleSidebarChange: function tzpf_handleSidebarChange(site, value) {
    cpref_log("handle sidebar change; new zoom: " + value);

    // Apply the pref to the browser.
    this.zoomManager.textZoom = value;

    // Save the new value to the datastore.
    this._cps.setPref(site, this.key, value);
  },

  handleMenuChange: function tzpf_handleMenuChange(value) {
    cpref_log("handle menu change; new zoom: " + value);

    // We don't have to apply the pref to the browser because the menu command
    // handler has done that already.

    // Update the control in the sidebar to reflect the change.
    this.updateControl(value);

    // Save the new value to the datastore.
    var site = gBrowser.mCurrentBrowser.currentURI.host;
    this._cps.setPref(site, this.key, value);
  },

  /**
   * Update the control for this preference in the sidebar, if the sidebar
   * is open at the moment.
   *
   * We have to turn off change-based events for the duration of the update
   * to prevent our change from being misinterpreted as a user change, so we
   * set the control's "updating" attribute, and change event handlers look
   * for it and return early without doing anything when they see it.
   *
   */
  updateControl: function tzpf_updateControl(value) {
    if (!this.sidebar)
      return;

    var control = this.sidebar.getElementById(this.key);

    if (!control)
      return;

    control.setAttribute("updating", true);
    control.value = value;
    // The change event fires now, but the event handlers see the attribute
    // and ignore the event.
    control.removeAttribute("updating");
  }
};

function cpref_tzpf_init() {
  gTextZoomPrefHandler.init();
  gContentPrefsController.registerPrefHandler(gTextZoomPrefHandler);
}

function cpref_tzpf_destroy() {
  gContentPrefsController.unregisterPrefHandler(gTextZoomPrefHandler);
  gTextZoomPrefHandler = null;
}

window.addEventListener("load", cpref_tzpf_init, false);
window.addEventListener("unload", cpref_tzpf_destroy, false);
